<?php

namespace Drupal\cmrf_form_processor\Plugin\WebformHandler;

use CMRF\Core\Call as CallInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\file\Entity\File;
use Drupal\webform\Plugin\WebformElement\BooleanBase;
use Drupal\webform\Plugin\WebformHandlerBase;
use Drupal\webform\Utility\WebformElementHelper;
use Drupal\webform\Utility\WebformOptionsHelper;
use Drupal\webform\WebformSubmissionConditionsValidatorInterface;
use Drupal\webform\WebformSubmissionInterface;
use Drupal\webform\WebformTokenManagerInterface;
use Drupal\cmrf_core\Core;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Serialization\Yaml;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;

/**
 * Webform submission debug handler.
 *
 * @WebformHandler(
 *   id = "cmrf_form_processor",
 *   label = @Translation("CiviCRM Form Processor with CiviMcRestFace (CMRF)"),
 *   category = @Translation("CiviCRM"),
 *   description = @Translation("Post values to CiviCRM form processor"),
 *   cardinality = \Drupal\webform\Plugin\WebformHandlerInterface::CARDINALITY_SINGLE,
 *   results = \Drupal\webform\Plugin\WebformHandlerInterface::RESULTS_PROCESSED,
 *   submission = \Drupal\webform\Plugin\WebformHandlerInterface::SUBMISSION_OPTIONAL,
 * )
 */
class FormProcessorWebformHandler extends WebformHandlerBase {

  /**
   * The webform token manager.
   *
   * @var \Drupal\webform\WebformTokenManagerInterface
   */
  protected $tokenManager;

  /**
   * The CiviMRF Core.
   *
   * @var \CMRF\Core\Core
   */
  protected $core;


  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $instance = parent::create($container, $configuration, $plugin_id, $plugin_definition);
    $instance->configFactory = $container->get('config.factory');
    $instance->conditionsValidator =   $container->get('webform_submission.conditions_validator');
    $instance->tokenManager =  $container->get('webform.token_manager');
    $instance->loggerFactory  =  $container->get('logger.factory');
    $instance->core = $container->get('cmrf_core.core');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function getSummary() {
    $t_args = ['%plugin_id' => $this->getPluginId(),
               '%connection' => $this->configuration['connection'],
               '%form_processor' =>  $this->configuration['form_processor'],
               '%form_processor_params' =>  '['.implode(',',$this->configuration['form_processor_params']).']',
    ];
    return [
      'message' => [
        '#markup' => $this->t('This %plugin_id handler using connection %connection and form_processor %form_processor with parameters %form_processor_params ', $t_args),
      ],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'connection' => null,
      'states' => [WebformSubmissionInterface::STATE_COMPLETED],
      'form_processor' => null,
      'form_processor_params' => [],
      'form_processor_current_contact' => 0,
      'form_processor_background' => 0,
      'form_processor_enable_validation' => 1,
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
   // $results_disabled = $this->getWebform()->getSetting('results_disabled');

    //parent::submitConfigurationForm($form, $form_state);
    $this->applyFormStateToConfiguration($form_state);
    $builder = new FormProcessorWebformBuilder($this->getWebform());

    $selected_connection = empty($this->configuration['connection'])?null:$this->configuration['connection'];
    $selected_formprocessor = empty($this->configuration['form_processor'])?null:$this->configuration['form_processor'];

    $form['fp'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('CMRF Form Processor'),
      '#attributes' => ['id' => 'form-processor-stuff'],
    ];
    $form['fp']['connection'] = [
      '#type' => 'select',
      '#title' => $this->t('Connector'),
      '#empty_option' => $this->t('- None -'),
      '#options' => $this->core->getConnectors(),
      '#default_value' => $this->configuration['connection'],
      '#ajax' => [
        'callback' => [$this, 'connectionCallback'],
        'wrapper' => 'form-processor-stuff',
        'event'   => 'change',
        ]
      ];
    if($selected_connection) {
      $form['fp']['form_processor'] = [
        '#type' => 'select',
        '#title' => $this->t('Form Processor'),
        '#empty_option' => $this->t('- None -'),
        '#options' => $this->formProcessorList($selected_connection),
        '#default_value' => $this->configuration['form_processor'],
      ];
      $form['additional'] =
        [ '#type' => 'fieldset',
          '#title' => $this->t('Form Processor'),
        ];

      // Settings: States.
      $results_disabled = $this->getWebform()->getSetting('results_disabled');
      $form['additional']['states'] = [
        '#type' => 'checkboxes',
        '#title' => $this->t('Send to CiviCRM'),
        '#options' => [
          WebformSubmissionInterface::STATE_DRAFT_CREATED => $this->t('…when <b>draft is created</b>.'),
          WebformSubmissionInterface::STATE_DRAFT_UPDATED => $this->t('…when <b>draft is updated</b>.'),
          WebformSubmissionInterface::STATE_CONVERTED => $this->t('…when anonymous <b>submission is converted</b> to authenticated.'),
          WebformSubmissionInterface::STATE_COMPLETED => $this->t('…when <b>submission is completed</b>.'),
          WebformSubmissionInterface::STATE_UPDATED => $this->t('…when <b>submission is updated</b>.'),
          WebformSubmissionInterface::STATE_DELETED => $this->t('…when <b>submission is deleted</b>.'),
          WebformSubmissionInterface::STATE_LOCKED => $this->t('…when <b>submission is locked</b>.'),
        ],
        '#access' => $results_disabled ? FALSE : TRUE,
        '#default_value' => $results_disabled ? [WebformSubmissionInterface::STATE_COMPLETED] : $this->configuration['states'],
      ];

      if ($selected_formprocessor) {
        $fpFields = $this->formProcessorFields($selected_connection, $selected_formprocessor);
        $fpFields = $this->mapTitle($fpFields);
        if (count($fpFields)) {
          $form['additional']['fields'] = [
            '#type' => 'fieldset',
            '#title' => $this->t('Fields'),
          ];
          $form['additional']['fields']['update_fields'] = [
            '#type' => 'select',
            '#title' => 'Update fields (this might delete the fields from your form)',
            '#default_value' => $this->configuration['form_processor_update_fields'] ?? 1,
            '#options' => ['0' => $this->t("No"), '1' => $this->t("Yes")],
          ];
          $form['additional']['fields']['help'] = [
            '#type' => 'markup',
            '#markup' => $this->t(
              'With Yes selected, saving this configuration will synchronise the Webform with the Form Processor input fields, adding (not updating) checked fields to and deleting unchecked fields from the Webform.'
            ),
          ];

          // Add a table of form processor fields currently in the form for
          // defining their format to be used for submissions to CiviCRM.
          $form['additional']['submission_settings'] = [
            '#type' => 'fieldset',
            '#title' => $this->t('Submission settings'),
            'fields' => [
              '#type' => 'table',
              '#header' => [
                'field' => ['data' => $this->t('Field'), 'width' => '50%'],
                'format' => [
                  'data' => $this->t('Format'),
                  'class' => [RESPONSIVE_PRIORITY_LOW],
                  'width' => '50%',
                ],
              ],
              '#sticky' => TRUE,
            ],
          ];
          /** @var \Drupal\webform\Plugin\WebformElementManagerInterface $element_manager */
          $element_manager = \Drupal::service('plugin.manager.webform.element');

          foreach ($fpFields as $key => $field) {
            // Add checkboxes for adding/deleting form processor fields to/from
            // the form when saving configuration.
            $form['additional']['fields']['field'][$key] = [
              '#type' => 'checkbox',
              '#title' => $field,
              '#default_value' => $builder->isFieldInForm($key) ? '1' : '0',
              '#states' => [
                'visible' => [
                  [
                    [':input[name="settings[additional][fields][update_fields]"]' => ['value' => '1']],
                  ]
                ]
              ]
            ];

            // Add configuration option for selecting the format to use for
            // sending field values to CiviCRM.
            if ($element = $this->getWebform()->getElement($key)) {
              $webform_element = $element_manager->getElementInstance($element);
              $form['additional']['submission_settings']['fields'][$key]['field'] = [
                '#markup' => $element['#title'],
              ];
              $form['additional']['submission_settings']['fields'][$key]['format'] = [
                '#type' => 'select',
                '#title' => $this->t('Item format'),
                '#description' => $this->t('Select how a single value is submitted to CiviCRM.'),
                '#options' => WebformOptionsHelper::appendValueToText($webform_element->getItemFormats()),
                // Default value for unconfigured elements with options or of a
                // Boolean type should be "Raw value" for sending option keys
                // instead of labels.
                '#default_value' =>
                  $this->configuration['submission_settings']['fields'][$key]['format']
                  ?? (
                    (
                      $webform_element->hasProperty('options')
                      || is_a($webform_element, BooleanBase::class)
                    ) ? 'raw'
                      : NULL
                  ),
              ];
            }
          }
        }

        $parameters = $this->formProcessorDefaultsParams($selected_connection, $selected_formprocessor);
        if (count($parameters)) {
          $form['additional']['params'] =
            [
              '#type' => 'fieldset',
              '#title' => $this->t('Parameters for retrieving default values'),
            ];
          foreach ($parameters as $key => $field) {
            $form['additional']['params']['param_'.$key] = [
              '#type' => 'select',
              '#title' => $field,
              '#default_value' => $this->configuration['form_processor_params'][$key],
              '#options' => [
                'none' => 'None',
                'url' => $this->t('URL'),
                'current_user' => $this->t('Current User'),
                'value' => $this->t('Value (you can use tokens)'),
              ]
            ];
            $form['additional']['params']['value_'.$key] = [
              '#type' => 'textfield',
              '#title' => $field,
              '#default_value' => $this->configuration['form_processor_params_value'][$key],
              '#states' => [
                'visible' => [
                  ':input[name="settings[additional][params][param_'.$key.']"]' => ['value' => 'value'],
                ],
                'required' => [
                  ':input[name="settings[additional][params][param_'.$key.']"]' => ['value' => 'value'],
                ],
              ],
            ];
          }
        }
        $form['additional']['params']['enable_validation'] = [
          '#type' => 'select',
          '#title' => 'Enable Validation with the Form Processor',
          '#default_value' => $this->configuration['form_processor_enable_validation'],
          '#options' => [
            0 => "No",
            1 => "Yes",
          ]
        ];
        $form['additional']['params']['require_default'] = [
          '#type' => 'select',
          '#title' => 'Require Default Values to be populated',
          '#default_value' => $this->configuration['form_processor_require_default'],
          '#options' => [
            0 => "No",
            1 => "Yes",
          ]
        ];
        if (count($fpFields)) {
          $form['additional']['form_processor_current_contact'] = [
            '#type' => 'select',
            '#title' => 'Fill Current Contact',
            '#default_value' => $this->configuration['form_processor_current_contact'],
            '#options' => [0 => "-None-"] + $fpFields
          ];
        }
      }
    }

    $this->elementTokenValidate($form);

    return $this->setSettingsParents($form);
  }

   public function connectionCallback(array $form, FormStateInterface $form_state) {
    return $form['settings']['fp'];
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {

    parent::submitConfigurationForm($form, $form_state);
    $this->applyFormStateToConfiguration($form_state);
    $values = $form_state->getValues();

    // Cleanup states.
    $values['states'] = array_values(array_filter($values['states']));
    $this->configuration['connection']            = $values['connection'];
    $this->configuration['states']                = $values['states'];
    $this->configuration['form_processor']        = $values['form_processor'];
    $this->configuration['form_processor_params'] = [];
    $this->configuration['form_processor_params_value'] = [];
    foreach($values['additional']['params'] as $key => $value) {
      if (stripos($key, 'param_')===0) {
        $this->configuration['form_processor_params'][substr($key, 6)] = $value;
      } elseif (stripos($key, 'value_')===0) {
        $this->configuration['form_processor_params_value'][substr($key, 6)] = $value;
      }
    }
    $this->configuration['form_processor_enable_validation'] = $values['additional']['params']['enable_validation'];
    $this->configuration['form_processor_require_default'] = $values['additional']['params']['require_default'];
    $this->configuration['form_processor_current_contact'] = $values['form_processor_current_contact'];
    if ($values['additional']['fields']['update_fields']) {
      $builder = new FormProcessorWebformBuilder($this->getWebform());
      $fpValues = $this->formProcessorFields($this->configuration['connection'], $this->configuration['form_processor']);
      $builder->addFields($values['additional']['fields']['field'], $fpValues);
      $builder->deleteFields($values['additional']['fields']['field']);
      $builder->save();
    }
    $this->configuration['form_processor_update_fields'] = $values['additional']['fields']['update_fields'];
    $this->configuration['submission_settings'] = $values['additional']['submission_settings'];
  }

  private function getContactId() {
    $id =  \Drupal::currentUser()->id();
    $currentUser = \Drupal::entityTypeManager()->getStorage('user')->load($id);
    return $currentUser->field_user_contact_id->value;
  }

  public function alterForm(array &$form, FormStateInterface $form_state, WebformSubmissionInterface $webform_submission) {
    $form_processor_params = $this->configuration['form_processor_params'];
    $form_processor_params_value = $this->configuration['form_processor_params_value'] ?? [];
    if(empty($form_processor_params)){
      if (!empty($this->configuration['form_processor_require_default'])) {
        throw new NotFoundHttpException();
      }
      return; // no form_processor params means no form defaults so return
    }
    $params = [];
    foreach($form_processor_params as $key => $fp_param){
      if($fp_param=='url' && \Drupal::request()->get($key)){
        $params[$key] = \Drupal::request()->get($key);
      }
      if($fp_param=='current_user' && $this->getContactId()){
        $params[$key] = $this->getContactId();
      }
      if ($fp_param=='value' && !empty($form_processor_params_value[$key])) {
        $parsedValue = \Drupal::token()->replace($form_processor_params_value[$key]);
        if (!empty($parsedValue)) {
          $params[$key] = $parsedValue;
        }
      }
    }
    if(empty($params)){
      if (!empty($this->configuration['form_processor_require_default'])) {
        throw new NotFoundHttpException();
      }
      return;
    }

    $values = $this->formProcessorDefaultsDefault($params);
    if($values['is_error']==0){
      unset($values['is_error']);
      foreach ($values as $key => $value) {
        $element =& WebformElementHelper::getElement($form, $key);
        if ($element) {
          if ($element['#type'] == 'datetime') {
            try {
              $element['#default_value'] = DrupalDateTime::createFromFormat(\DateTime::ATOM, $value);
            } catch (\InvalidArgumentException $ex) {
              // is it something else than a date ignore
            }
          }
          else {
            $element['#default_value'] = $value;
          }
        }
      }
    }
    else {
      if (!empty($this->configuration['form_processor_require_default'])) {
        throw new NotFoundHttpException();
      }
    }
  }


  /**
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   * @param \Drupal\webform\WebformSubmissionInterface $webform_submission
   *
   * @return void
   */
  public function validateForm(array &$form, FormStateInterface $form_state, WebformSubmissionInterface $webform_submission) {
    if (empty($this->configuration['form_processor_enable_validation'])) {
      return;
    }
    $params = $this->webformSubmissionToApiParams($webform_submission);
    $call = $this->core->createCall(
      $this->configuration['connection'],
      'FormProcessorValidation',
      $this->configuration['form_processor'],
      $params,
      []
    );
    $this->core->executeCall($call);
    if ($call->getStatus() == CallInterface::STATUS_DONE) {
      foreach ($call->getReply() as $field => $msg) {
        if ($field == 'is_error') {
          continue;
        }
        $form_state->setErrorByName($field, $msg);
      }
    }
  }

  /**
   * {@inheritdoc}
   * @param \Drupal\webform\WebformSubmissionInterface $webform_submission
   * @param $update
   * @param $extraParams array used for example to add the results of a payment provider.
   *
   * @return void
   */
  public function postSave(WebformSubmissionInterface $webform_submission, $update = TRUE, $extraParams = []) {
    // $extraParams might be passed as NULL.
    $extraParams ??= [];
    $state = $webform_submission->getWebform()->getSetting('results_disabled') ? WebformSubmissionInterface::STATE_COMPLETED : $webform_submission->getState();
    if ($this->configuration['states'] && in_array($state, $this->configuration['states'])) {
      $this->sendToCiviCRM($webform_submission, $extraParams);
    }
  }

  /**
   * {@inheritDoc}
   */
  public function postDelete(WebformSubmissionInterface $webform_submission) {
    if (in_array(WebformSubmissionInterface::STATE_DELETED, $this->configuration['states'])) {
      $this->sendToCiviCRM($webform_submission);
    }
  }


  /**
   * @param \Drupal\webform\WebformSubmissionInterface $webform_submission
   * @param $extraParams array used for example to add the results of a payment provider.
   *
   * @return void
   */
  private function sendToCiviCRM(WebformSubmissionInterface $webform_submission, $extraParams = []) {
    $params = $this->webformSubmissionToApiParams($webform_submission);
    $params = array_merge($params, $extraParams);
    $call = $this->core->createCall(
      $this->configuration['connection'],
      'FormProcessor',
      $this->configuration['form_processor'],
      $params,
      []
    );
    $this->core->executeCall($call);
  }

  private function formProcessorList($connection){
     $call = $this->core->createCall($connection,'FormProcessorInstance','list',[],['limit'=>0]);
     $values = $this->core->executeCall($call)['values'];
     return array_map(function($value) { return $value['title']; } ,$values);
  }

  private function formProcessorFields($connection,$formprocessor){
    $call = $this->core->createCall($connection,'FormProcessor','getfields',['api_action'=>$formprocessor],['limit'=>0]);
    return $this->core->executeCall($call)['values'];
  }

  private function mapTitle($values){
    return array_map(function($value) { return $value['title']; } ,$values);
  }

  private function formProcessorDefaultsParams($connection, $formprocessor){
    $call = $this->core->createCall($connection,'FormProcessorDefaults','getfields',['api_action'=>$formprocessor],['limit'=>0],[]);
    $result = $this->core->executeCall($call);
    if($result['count']==0){
      return [];
    } else {
      return array_map(function($value) { return $value['name']; } ,$result['values']);
    }
  }

  private function formProcessorDefaultsDefault($params){
    $call = $this->core->createCall($this->configuration['connection'],'FormProcessorDefaults',$this->configuration['form_processor'],$params,[],[]);
    $result = $this->core->executeCall($call);
    return $result;
  }

  private function webformSubmissionToApiParams(WebformSubmissionInterface $webform_submission): array {
    $data = $webform_submission->getData();
    $fields = $this->formProcessorFields($this->configuration['connection'], $this->configuration['form_processor']);
    foreach ($fields as $key => $field) {
      if (key_exists($key, $data)) {
        if (isset($data[$key]) && in_array($this->getWebform()
            ->getElement($key)['#type'], [
            'webform_document_file',
            'managed_file',
            'webform_image_file',
          ])) {
          $file = File::load($data[$key]);
          $fileData = [
            'name' => $file->getFilename(),
            'mime_type' => $file->getMimeType(),
            'content' => base64_encode(file_get_contents($file->getFileUri())),
          ];
          $params[$key] = $fileData;
        }
        elseif ($this->configuration['submission_settings']['fields'][$key]['format'] && $this->configuration['submission_settings']['fields'][$key]['format'] != 'raw') {
          // Use the selected format from configuration for rendering the field
          // value. We're using Webform's buildExportRecord() method for
          // rendering, as it takes care of element-specific formatting, such as
          // deciding whether to render as HTML or text.
          /** @var \Drupal\webform\Plugin\WebformElementManagerInterface $element_manager */
          $element_manager = \Drupal::service('plugin.manager.webform.element');
          $element = $this->getWebform()->getElement($key);
          $element['#format'] = $this->configuration['submission_settings']['fields'][$key]['format'];
          $export_record = $element_manager
            ->getElementInstance($element)
            ->buildExportRecord(
              $element,
              $webform_submission,
              ['webform' => $this->getWebform()]
            );
          $params[$key] = (string) reset($export_record);
        }
        else {
          $params[$key] = $data[$key];
        }
      }
    }
    if ($this->configuration['form_processor_current_contact']) {
      $params[$this->configuration['form_processor_current_contact']] = $this->getContactId();
    }
    return $params;
  }
}
